@@ -4,32 +4,44 @@ import os
 import re
 import subprocess
 from copy import copy
-from os.path import basename, dirname, exists, join, relpath, splitext
+from os.path import basename
+from os.path import dirname
+from os.path import exists
+from os.path import join
+from os.path import relpath
+from os.path import splitext
 from shutil import rmtree
-from typing import Any, Dict, List, cast
+from typing import Any
+from typing import Dict
+from typing import List
+from typing import cast
 
 from jinja2 import Template
 
 from dipdup import __version__
-from dipdup.config import (
-    BigMapIndexConfig,
-    CallbackMixin,
-    ContractConfig,
-    DatasourceConfigT,
-    DipDupConfig,
-    HandlerConfig,
-    HeadIndexConfig,
-    IndexTemplateConfig,
-    OperationHandlerOriginationPatternConfig,
-    OperationHandlerTransactionPatternConfig,
-    OperationIndexConfig,
-    TzktDatasourceConfig,
-    default_hooks,
-)
+from dipdup.config import BigMapIndexConfig
+from dipdup.config import CallbackMixin
+from dipdup.config import ContractConfig
+from dipdup.config import DatasourceConfigT
+from dipdup.config import DipDupConfig
+from dipdup.config import HandlerConfig
+from dipdup.config import HeadIndexConfig
+from dipdup.config import IndexTemplateConfig
+from dipdup.config import OperationHandlerOriginationPatternConfig
+from dipdup.config import OperationHandlerTransactionPatternConfig
+from dipdup.config import OperationIndexConfig
+from dipdup.config import TzktDatasourceConfig
+from dipdup.config import default_hooks
 from dipdup.datasources.datasource import Datasource
 from dipdup.datasources.tzkt.datasource import TzktDatasource
-from dipdup.exceptions import ConfigInitializationException, ConfigurationError
-from dipdup.utils import import_submodules, mkdir_p, pascal_to_snake, snake_to_pascal, touch, write
+from dipdup.exceptions import ConfigInitializationException
+from dipdup.exceptions import ConfigurationError
+from dipdup.utils import import_submodules
+from dipdup.utils import mkdir_p
+from dipdup.utils import pascal_to_snake
+from dipdup.utils import snake_to_pascal
+from dipdup.utils import touch
+from dipdup.utils import write
 
 DEFAULT_DOCKER_ENV_FILE_CONTENT = dict(
     POSTGRES_USER="dipdup",
@@ -49,13 +61,25 @@ DEFAULT_DOCKER_ENV_FILE = 'dipdup.env'
 _templates: Dict[str, Template] = {}
 
 
-def resolve_big_maps(schema: Dict[str, Any]) -> Dict[str, Any]:
+def preprocess_storage_jsonschema(schema: Dict[str, Any]) -> Dict[str, Any]:
     """Preprocess bigmaps in JSONSchema. Those are unions as could be pointers.
     We resolve bigmaps from diffs so no need to include int in type signature."""
+    if not isinstance(schema, dict):
+        return schema
     if 'properties' in schema:
         return {
             **schema,
-            'properties': {prop: resolve_big_maps(sub_schema) for prop, sub_schema in schema['properties'].items()},
+            'properties': {prop: preprocess_storage_jsonschema(sub_schema) for prop, sub_schema in schema['properties'].items()},
+        }
+    elif 'items' in schema:
+        return {
+            **schema,
+            'items': preprocess_storage_jsonschema(schema['items']),
+        }
+    elif 'additionalProperties' in schema:
+        return {
+            **schema,
+            'additionalProperties': preprocess_storage_jsonschema(schema['additionalProperties']),
         }
     elif schema.get('$comment') == 'big_map':
         return schema['oneOf'][1]
@@ -153,8 +177,8 @@ class DipDupCodeGenerator:
                         mkdir_p(contract_schemas_path)
 
                         storage_schema_path = join(contract_schemas_path, 'storage.json')
+                        storage_schema = preprocess_storage_jsonschema(contract_schemas['storageSchema'])
 
-                        storage_schema = resolve_big_maps(contract_schemas['storageSchema'])
                         write(storage_schema_path, json.dumps(storage_schema, indent=4, sort_keys=True))
 
                         if not isinstance(operation_pattern_config, OperationHandlerTransactionPatternConfig):
@@ -267,7 +291,6 @@ class DipDupCodeGenerator:
                     '--class-name',
                     name.lstrip('_'),
                     '--disable-timestamp',
-                    '--use-default',
                 ]
                 self._logger.debug(' '.join(args))
                 subprocess.run(args, check=True)
@@ -305,15 +328,15 @@ class DipDupCodeGenerator:
         write(join(docker_path, 'Dockerfile'), dockerfile_code, overwrite=True)
 
         mounts = {}
-        for filename in self._config.filenames:
-            filename_part = filename.split("/")[-1]
-            from_ = join(relpath(self._config.package_path, filename), filename_part)
-            to = f'/home/dipdup/{filename_part}'
+        for path in self._config.paths:
+            path_part = path.split("/")[-1]
+            from_ = join(relpath(self._config.package_path, path), path_part)
+            to = f'/home/dipdup/{path_part}'
             mounts[from_] = to
 
         command = []
-        for filename in self._config.filenames:
-            command += ['-c', filename.split("/")[-1]]
+        for path in self._config.paths:
+            command += ['-c', path.split("/")[-1]]
         command += ['run']
 
         docker_compose_code = docker_compose_template.render(
